(* ========================================================================= *)
(*                                                                           *)
(*        		   Application.ml					      *)
(*                                                                           *)
(*        (c) Copyright, Sidi Mohamed BEILLAHI, Muhammad Umair Siddique      *)
(*                       2015                                                *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <beillahi@encs.concordia.ca>,				     *)
(*		<muh_sidd@ece.concordia.ca>                                  *)
(*       	                                 			     *)
(*                                                                           *)
(* Last update: March 02, 2015                                               *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)



needs "Stability_Resonance.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)
(*-------------------------------------------------------------------*)
(*-----------------1-Boost Cell DC-DC COnverter-----------------------*)
(*-------------------------------------------------------------------*)

let cell1_help1 =
 prove (`
~((r1+s*L1)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))  ==> 
(Cx (&1) +
      Cx (&0) +
      Cx (&0) +
      Cx (&0) -
      Cx (&1) / (s * L1 + r1) * D1 * R / (Cx (&1) + s * R * C) * --D1) = 
	  ((r1 + s * L1) * (Cx (&1) + s * R * C) + R * D1 pow 2) / 
	  ((r1 + s * L1) * (Cx (&1) + s * R * C))`, CONV_TAC COMPLEX_FIELD);;


let cell1_circuit1 = prove ( 
 `~((r1+s*L1)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) ==> 
SFG_Main ([1,Cx(&1),2;2,Cx(&1)/(s*L1+r1),3;3,Cx(&1),4; 3,D1,5;
5,R/(Cx(&1)+s*R*C),6;6,--D1,2], 6, 4, 1) = 
(Cx(&1)+s*R*C)/((r1+s*L1)*(Cx(&1)+s*R*C)+R* D1 pow 2) `,
SFG_TAC THEN SIMP_TAC [cell1_help1] THEN CONV_TAC COMPLEX_FIELD);; 

 let cell1_circuit2 = prove ( 
 `~((r1+s*L1)=Cx(&0)) /\  ~((Cx(&1)+s*R*C)=Cx(&0))  ==> 
SFG_Main ([1,Cx(&1),2;2,Cx(&1)/(s*L1+r1),3; 3,Cx(&1),4;3,D1,5;
5,R/(Cx(&1)+s*R*C),6;6,--D1,2], 6, 6, 1) = 
(R* D1)/((r1+s*L1)*(Cx(&1)+s*R*C)+R*D1 pow 2) `,
SFG_TAC THEN SIMP_TAC [cell1_help1] THEN CONV_TAC COMPLEX_FIELD);; 


let cell1_circuit3 = prove ( 
`~((r1+s*L1)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))  ==> 
SFG_Main ([1,Cx(&1),2;2,Cx(&1)/(s*L1+r1),3;3,Cx(&1),4; 3,D1,5;
5,R/(Cx(&1)+s*R*C),6;6,--D1,2], 6, 5, 1) = 
(Cx(&1)+s*R*C)*D1/((r1+s*L1)*(Cx(&1)+s*R*C)+R* D1 pow 2)`,
SFG_TAC THEN SIMP_TAC [cell1_help1] THEN CONV_TAC COMPLEX_FIELD);; 

(*----------------------------EFFICIENCY-----------------------------------*)

let cell1_efficiency = prove (`
~((r1+s*L1)=Cx(&0))  /\ ~((Cx(&1)+s*R*C)=Cx(&0)) 
/\ ~(((r1+s*L1)*(Cx(&1)+s*R*C)+R* D1 pow 2)=Cx(&0))   ==> 
(SFG_Main ([1,Cx(&1),2;2,Cx(&1)/(s*L1+r1),3; 3,Cx(&1),4;3,D1,5;
5,R/(Cx(&1)+s*R*C),6;6,--D1,2], 6, 6, 1))* (SFG_Main 
([1,Cx(&1),2;2,Cx(&1)/(s*L1+r1),3;3,Cx(&1),4; 3,D1,5;
5,R/(Cx(&1)+s*R*C),6;6,--D1,2], 6, 5, 1))/(SFG_Main 
([1,Cx(&1),2;2,Cx(&1)/(s*L1+r1),3;3,Cx(&1),4; 3,D1,5;
5,R/(Cx(&1)+s*R*C),6;6,--D1,2], 6, 4, 1)) = 
(R * D1 pow 2 )/((r1+s*L1)*(Cx(&1)+s*R*C)+R* D1 pow 2)`, 
SIMP_TAC [cell1_circuit1;cell1_circuit2;cell1_circuit3] THEN
CONV_TAC COMPLEX_FIELD);;

(*----------------------------Stability-------------------------------------*)

let Cell1_SFG = define `Cell1_SFG L1 r1 R C D1 s = 
([1,Cx(&1),2;2,Cx(&1)/(s*L1+r1),3; 3,Cx(&1),4;3,D1,5;
5,R/(Cx(&1)+s*R*C),6;6,--D1,2], 6, 6, 1)`;;

(*--------------------------------------------------------------------*)

let DENOM_OF_cell1 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))  
	  ==> DENOMINATOR (Cell1_SFG L1 r1 R C D1 s) =  
	  ((R*C*L1)*s pow 2 + (R*C*r1+L1)*s + (r1+R* D1 pow 2))/((r1+s*L1)*(Cx(&1)+s*R*C))`,
SIMP_TAC [Cell1_SFG] THEN
DENOMINATOR_TAC
THEN CONV_TAC COMPLEX_FIELD);;
(*-----------------------------------------------------------*)
let help0 = prove( ` !z a1 a2 a0. 
z = (--a1 + csqrt (a1 pow 2 - (Cx(&4)) * a2 * a0))/((Cx(&2)) * a2) 
/\ ~(a2 = Cx (&0)) ==> a2 * z pow 2 + a1 * z + a0 = Cx (&0)`,
REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] 
THEN SIMP_TAC[COMPLEX_POW_2;COMPLEX_POW_DIV] THEN 
SIMP_TAC [COMPLEX_ADD_LDISTRIB;COMPLEX_ADD_RDISTRIB] 
THEN REWRITE_TAC[GSYM COMPLEX_POW_2;COMPLEX_POW_INV;CSQRT] 
THEN REPEAT(POP_ASSUM MP_TAC) THEN  TRY(CONV_TAC COMPLEX_FIELD));;


let cell1_help2 = prove( `  
s = (--(R*C*r1+L1)+ csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1) /\ ~((R*C*L1) = Cx (&0)) ==> 
(R*C*L1)*s pow 2 + (R*C*r1+L1)*s + (r1+R* D1 pow 2) = Cx (&0)`,SIMP_TAC[help0] );;

let help01 = prove( ` !z a1 a2 a0. 
z = (--a1 - csqrt (a1 pow 2 - (Cx(&4)) * a2 * a0))/((Cx(&2)) * a2) 
/\ ~(a2 = Cx (&0)) ==> a2 * z pow 2 + a1 * z + a0 = Cx (&0)`,
REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
SIMP_TAC[COMPLEX_POW_2;COMPLEX_POW_DIV] THEN 
SIMP_TAC [COMPLEX_SUB_LDISTRIB;COMPLEX_SUB_RDISTRIB] THEN 
REWRITE_TAC[GSYM COMPLEX_POW_2;COMPLEX_POW_INV;CSQRT] THEN 
REPEAT(POP_ASSUM MP_TAC) THEN  TRY(CONV_TAC COMPLEX_FIELD));;


let cell1_help3 = prove( `  
s = (--(R*C*r1+L1) - csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1) /\ ~((R*C*L1) = Cx (&0)) ==> 
(R*C*L1)*s pow 2 + (R*C*r1+L1)*s + (r1+R* D1 pow 2) = Cx (&0)`,SIMP_TAC[help01] );;


let cell1_help4 = prove( `  ~((R*C*L1) = Cx (&0)) ==> 
	((R * C * L1) *
      ((--(R * C * r1 + L1) +
        csqrt
   ((R * C * r1 + L1) pow 2 - Cx (&4) * (R * C * L1) * (r1 +R* D1 pow 2))) /
       (Cx (&2) * R * C * L1)) pow
      2 +
      (R * C * r1 + L1) *
      (--(R * C * r1 + L1) +
       csqrt
   ((R * C * r1 + L1) pow 2 - Cx (&4) * (R * C * L1) * (r1 + R*D1 pow 2))) /
      (Cx (&2) * R * C * L1) +
      r1 +
      R * D1 pow 2) = Cx (&0) /\
((R * C * L1) *
      ((--(R * C * r1 + L1) -
        csqrt
        ((R * C * r1 + L1) pow 2 - Cx (&4) * (R * C * L1) * (r1 + R*D1 pow 2))) /
       (Cx (&2) * R * C * L1)) pow
      2 +
      (R * C * r1 + L1) *
      (--(R * C * r1 + L1) -
       csqrt
       ((R * C * r1 + L1) pow 2 - Cx (&4) * (R * C * L1) * (r1 + R*D1 pow 2))) /
      (Cx (&2) * R * C * L1) +
      r1 +
      R * D1 pow 2) = Cx (&0)`,MESON_TAC[cell1_help2;cell1_help3] );;

let cell1_stability_conditions = prove(  
`(norm ((--(R*C*r1+L1) + csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1) ) <= &1) 
/\ (norm ((--(R*C*r1+L1) - csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2)))
 / ((Cx(&2))*R*C*L1) ) <= &1) /\ ~(R*C*L1=Cx(&0)) /\  
 ~((--(R*C*r1+L1) + csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
 / ((Cx(&2))*R*C*L1) =Cx(&0)) 
/\  ~(r1 + ((--(R*C*r1+L1) - csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1)) *L1 =Cx(&0)) 
/\  ~(r1 + ((--(R*C*r1+L1) + csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1)) * L1 =Cx(&0)) 
/\  ~(Cx(&1) + ((--(R*C*r1+L1) - csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1)) *R*C =Cx(&0)) 
/\  ~(Cx(&1) + ((--(R*C*r1+L1) + csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1)) *R*C =Cx(&0)) 
/\  ~((--(R*C*r1+L1) - csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
/ ((Cx(&2))*R*C*L1) =Cx(&0)) 
	==> is_stable (\s. Cell1_SFG L1 r1 R C D1 s) s 
	[(--(R*C*r1+L1) + csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2))) 
	/ ((Cx(&2))*R*C*L1);
	(--(R*C*r1+L1) - csqrt ((R*C*r1+L1) pow 2 - (Cx(&4))*(R*C*L1)*(r1+R*D1 pow 2)))
	/ ((Cx(&2))*R*C*L1)]`,
REWRITE_TAC [is_stable;poles_in_unit_circle;
	are_valid_poles;in_unit_circle;
	is_valid_pole;set_of_poles] THEN 
	REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC) THEN
REWRITE_TAC[IN_ELIM_THM] THEN
SIMP_TAC[DENOM_OF_cell1] THEN
SIMP_TAC[cell1_help4] THEN
CONV_TAC COMPLEX_FIELD THEN
REWRITE_TAC[IN_ELIM_THM] THEN
SIMP_TAC[DENOM_OF_cell1] THEN
SIMP_TAC[cell1_help4]
THEN CONV_TAC COMPLEX_FIELD);;	  
	
(*----------------------------------------------------------------*) 
(*-----------------2-Boost Cells Converter    --------------------*)  
(*----------------------------------------------------------------*) 


let cell2_help1 = prove (
`~((r1+s*L1) = Cx (&0)) /\ ~((r2+s*L2) = Cx (&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))
 ==> ((Cx (&1) *
       Cx (&1) / (r1 + s * L1) *
       D2 *
       R / (Cx (&1) + s * R * C) *
       --D1 *
       Cx (&1) / (r2 + s * L2) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r1 + s * L1) * Cx (&1)) *
      (Cx (&1) +
       Cx (&0) +
       Cx (&0) +
       Cx (&0) -
       Cx (&1) / (r2 + s * L2) * D1 * R / (Cx (&1) + s * R * C) * --D1) +
      (Cx (&1) * Cx (&1) / (r2 + s * L2) * Cx (&1)) *
      (Cx (&1) +
       Cx (&0) +
       Cx (&0) +
       Cx (&0) -
       Cx (&1) / (r1 + s * L1) * D2 * R / (Cx (&1) + s * R * C) * --D2) +
      (Cx (&1) *
       Cx (&1) / (r2 + s * L2) *
       D1 *
       R / (Cx (&1) + s * R * C) *
       --D2 *
       Cx (&1) / (r1 + s * L1) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      Cx (&0))  = (((r1+s*L1)+(r2+s*L2))*(Cx(&1)+s*R*C)+R* (D1-D2) pow 2 )
	  /((r1+s*L1)*(r2+s*L2)*(Cx(&1)+s*R*C))`, CONV_TAC COMPLEX_FIELD);;


let cell2_help2 = prove (`
~((r1+s*L1) = Cx (&0)) /\ ~((r2+s*L2) = Cx (&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))
 ==> (Cx (&1) +
      Cx (&0) +
      Cx (&0) +
      (Cx (&0) + Cx (&0) + Cx (&0) - Cx (&1) 
	  / (r2+s*L2) * D1 * R/(Cx(&1)+s*R*C) * --D1) -
      Cx (&1) / (r1+s*L1) * D2 * R/(Cx(&1)+s*R*C) * --D2) = 
	  ((Cx(&1)+s*R*C)+ R*(D1 pow 2 / (r2+s*L2) + D2 pow 2 / (r1+s*L1)))
	  /(Cx(&1)+s*R*C)`, CONV_TAC COMPLEX_FIELD);;

	  
let cell2_circuit1 = 
    prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) 
	/\  ~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r2+s*L2) + D2 pow 2 / (r1+s*L1))) = Cx (&0)) ==> 
SFG_Main 
([1,Cx(&1),2;1,Cx(&1),4;2,Cx(&1)/(r1+s*L1),3;3,D2,7; 3,Cx(&1),5;4,Cx(&1)/
(r2+s*L2),6;6,Cx(&1),5;6,D1,7;7,R/(Cx(&1)+s*R*C),8;8,--D1,4;8,--D2,2], 8, 5, 1) = 
(((r1+s*L1)+(r2+s*L2))*(Cx(&1)+s*R*C)+R* (D1-D2) pow 2 )/
((r1+s*L1)*(r2+s*L2)*(Cx(&1)+s*R*C)+R*((r1+s*L1)* D1 pow 2 + (r2+s*L2) * D2 pow 2)) `,
SFG_TAC
THEN SIMP_TAC [cell2_help1;cell2_help2]
THEN CONV_TAC COMPLEX_FIELD);; 


let cell2_help3 = prove (`~((r1+s*L1) = Cx (&0)) /\ 
~((r2+s*L2) = Cx (&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))
 ==> (R * ((r1 + s * L1) * D1 + (r2 + s * L2) * D2)) /
     ((r1 + s * L1) * (r2 + s * L2) * (Cx (&1) + s * R * C)) /
     (((Cx (&1) + s * R * C) +
       R * (D1 pow 2 / (r2 + s * L2) + D2 pow 2 / (r1 + s * L1))) /
      (Cx (&1) + s * R * C)) = 
(R * ((r1 + s * L1) * D1 + (r2 + s * L2) * D2)) /
    (((r1 + s * L1) * (r2 + s * L2)) * (((Cx (&1) + s * R * C) +
       R * (D1 pow 2 / (r2 + s * L2) + D2 pow 2 / (r1 + s * L1)))))`, 
CONV_TAC COMPLEX_FIELD);;



let cell2_help4 = prove (
`~((r1+s*L1) = Cx (&0)) /\ ~((r2+s*L2) = Cx (&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))
 ==> ((Cx (&1) * Cx (&1) / (r1+s*L1) * D2 * R/(Cx(&1)+s*R*C)) * (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r2+s*L2) * D1 * R/(Cx(&1)+s*R*C)) * (Cx (&1) + Cx (&0)) +
      Cx (&0)) = (R * ((r1+s*L1) * D1 + (r2+s*L2) * D2))/
	  ((r1+s*L1)*(r2+s*L2)*(Cx(&1)+s*R*C))`, CONV_TAC COMPLEX_FIELD);;



let cell2_help5 = prove (
`~((r1+s*L1) = Cx (&0)) /\ ~((r2+s*L2) = Cx (&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))
 ==> 
    ((r1 + s * L1) * (r2 + s * L2)) * (((Cx (&1) + s * R * C) +
       R * (D1 pow 2 / (r2 + s * L2) + D2 pow 2 / (r1 + s * L1)))) = 
	((r1+s*L1)*(r2+s*L2)*(Cx(&1)+s*R*C)+R*((r1+s*L1)* D1 pow 2 + (r2+s*L2) * D2 pow 2))`,
	CONV_TAC COMPLEX_FIELD);;


let cell2_circuit2 = 
      prove (
`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) 
/\  ~((Cx (&1) + R * (D1 pow 2 / r2 + D2 pow 2 / (r1+s*L1))) = Cx (&0)) ==> 
SFG_Main 
([1,Cx(&1),2;1,Cx(&1),4;2,Cx(&1)/(r1+s*L1),3;3,D2,7; 3,Cx(&1),5;4,Cx(&1)/
(r2+s*L2),6;6,Cx(&1),5;6,D1,7;7,R/(Cx(&1)+s*R*C),8;8,--D1,4;8,--D2,2], 8, 8, 1) = 
(R*((r1+s*L1)* D1 + (r2+s*L2) * D2))/
((r1+s*L1)*(r2+s*L2)*(Cx(&1)+s*R*C)+R*((r1+s*L1)* D1 pow 2 + (r2+s*L2) * D2 pow 2)) 
`, SFG_TAC 
THEN SIMP_TAC [cell2_help2;cell2_help4]
THEN SIMP_TAC [cell2_help3;cell2_help5]);; 


let cell2_help6 = prove (`~((r1+s*L1) = Cx (&0)) /\ ~((r2+s*L2) = Cx (&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0))
 ==> ((Cx (&1) * Cx (&1) / (r1+s*L1) * D2) * (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r2+s*L2) * D1) * (Cx (&1) + Cx (&0)) +
      Cx (&0)) = ( ((r1+s*L1) * D1 + (r2+s*L2) * D2))/((r1+s*L1)*(r2+s*L2))`, 
	  CONV_TAC COMPLEX_FIELD);;


let cell2_circuit3 = 
      prove (
`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) 
/\  ~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r2+s*L2) + D2 pow 2 / (r1+s*L1))) = Cx (&0)) ==> 
SFG_Main 
([1,Cx(&1),2;1,Cx(&1),4;2,Cx(&1)/(r1+s*L1),3;3,D2,7; 3,Cx(&1),5;4,Cx(&1)/
(r2+s*L2),6;6,Cx(&1),5;6,D1,7;7,R/(Cx(&1)+s*R*C),8;8,--D1,4;8,--D2,2], 8, 7, 1) = 
(Cx(&1)+s*R*C)*((r1+s*L1)* D1 + (r2+s*L2) * D2)/
((r1+s*L1)*(r2+s*L2)*(Cx(&1)+s*R*C)+R*((r1+s*L1)* D1 pow 2 + (r2+s*L2) * D2 pow 2)) 
`,
SFG_TAC
THEN SIMP_TAC [cell2_help2;cell2_help6]
THEN CONV_TAC COMPLEX_FIELD);; 

(*----------------------------------------------------------------*) 
(*-----------------3-Boost Cells Converter    --------------------*)  
(*----------------------------------------------------------------*) 
let cell3_help1 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\  
	  ~((r3+s*L3)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) /\
~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r3+s*L3) + D2 pow 2 / 
(r2+s*L2) + D3 pow 2 / (r1+s*L1))) = Cx (&0)) ==> 
((Cx (&1) * Cx (&1) / (r1+s*L1) * D3 * R/(Cx(&1)+s*R*C)) * (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r2+s*L2) * D2 * R/(Cx(&1)+s*R*C)) * (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r3+s*L3) * D1 * R/(Cx(&1)+s*R*C)) * (Cx (&1) + Cx (&0)) +
      Cx (&0)) = (R*((r2+s*L2)*(r3+s*L3)*D3  +
  (r1+s*L1)*(r3+s*L3)*D2 +(r1+s*L1)*(r2+s*L2)*D1)) 
  /((r1+s*L1)*(r2+s*L2)*(r3+s*L3)*(Cx(&1)+s*R*C))`, CONV_TAC COMPLEX_FIELD);;

let cell3_help2 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) 
	  /\  ~((r3+s*L3)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) /\
~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r3+s*L3) + D2 pow 2 / 
(r2+s*L2) + D3 pow 2 / (r1+s*L1))) = Cx (&0)) ==> 
(Cx (&1) +
Cx (&0) +
Cx (&0) +
(Cx (&0) +
 Cx (&0) +
(Cx (&0) + Cx (&0) + Cx (&0) - Cx (&1) / 
(r3+s*L3) * D1 * R/(Cx(&1)+s*R*C) * --D1) -
Cx (&1) / (r2+s*L2) * D2 * R/(Cx(&1)+s*R*C) * --D2) -
 Cx (&1) / (r1+s*L1) * D3 * R/(Cx(&1)+s*R*C) * --D3) = 
 ((Cx(&1)+s*R*C)+ R*(D1 pow 2 / (r3+s*L3) + D2 pow 2 / (r2+s*L2) +  D3 pow 2 / 
 (r1+s*L1)))/(Cx(&1)+s*R*C)`, CONV_TAC COMPLEX_FIELD);;




let cell3_help3 = 
      prove (`~(r1 + s * L1 = Cx (&0)) /\
 ~(r2 + s * L2 = Cx (&0)) /\
 ~(r3 + s * L3 = Cx (&0)) /\
 ~(Cx (&1) + s * R * C = Cx (&0)) /\
 ~((Cx (&1) + s * R * C) +
   R *
   (D1 pow 2 / (r3 + s * L3) +
    D2 pow 2 / (r2 + s * L2) +
    D3 pow 2 / (r1 + s * L1)) =
   Cx (&0))
 ==> 
(((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3)) *((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1)))) =    
((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3)*(Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 * (r1 + s * L1) * (r2 + s * L2)  +
        D2 pow 2 * (r1 + s * L1)  * (r3 + s * L3) +
        D3 pow 2 * (r2 + s * L2) * (r3 + s * L3)))  `, 
		CONV_TAC COMPLEX_FIELD);; 

		
let cell3_help4 = 
      prove (`~(r1 + s * L1 = Cx (&0)) /\
 ~(r2 + s * L2 = Cx (&0)) /\
 ~(r3 + s * L3 = Cx (&0)) /\
 ~(Cx (&1) + s * R * C = Cx (&0)) /\
 ~((Cx (&1) + s * R * C) +
   R *
   (D1 pow 2 / (r3 + s * L3) +
    D2 pow 2 / (r2 + s * L2) +
    D3 pow 2 / (r1 + s * L1)) =
   Cx (&0))
 ==> (R *
      ((r2 + s * L2) * (r3 + s * L3) * D3 +
       (r1 + s * L1) * (r3 + s * L3) * D2 +
       (r1 + s * L1) * (r2 + s * L2) * D1)) /
 ((r1 + s * L1) * (r2 + s * L2) * 
 (r3 + s * L3) * (Cx (&1) + s * R * C)) /
     (((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1))) /
      (Cx (&1) + s * R * C)) = (R *
      ((r2 + s * L2) * (r3 + s * L3) * D3 +
       (r1 + s * L1) * (r3 + s * L3) * D2 +
       (r1 + s * L1) * (r2 + s * L2) * D1)) /
     (((r1 + s * L1) * (r2 + s * L2) * 
	 (r3 + s * L3)) *((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1)))) `, CONV_TAC COMPLEX_FIELD);; 


let cell3_circuit1 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\  
	  ~((r3+s*L3)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) /\
~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r3+s*L3) + D2 pow 2 / 
(r2+s*L2) + D3 pow 2 / (r1+s*L1))) = Cx (&0)) ==> SFG_Main 
([1,Cx(&1),2;1,Cx(&1),4;1,Cx(&1),7;2,Cx(&1)/(r1+s*L1),3;
 3,Cx(&1),5;3,D3,9;4,Cx(&1)/(r2+s*L2),6;6,Cx(&1),5;6,D2,9;
 7,Cx(&1)/(r3+s*L3),8;8,D1,9;8,Cx(&1),5;9,R/(Cx(&1)+s*R*C),10;
 10,--D2,4;10,--D3,2;10,--D1,7], 10, 10,1) = 
(R*((r2+s*L2)*(r3+s*L3)*D3  +
  (r1+s*L1)*(r3+s*L3)*D2 +
  (r1+s*L1)*(r2+s*L2)*D1)) /
 ((r1+s*L1)*(r2+s*L2)*(r3+s*L3)*(Cx(&1)+s*R*C) +
  R*((r1+s*L1)*(r2+s*L2)*(D1 pow 2) +
	(r1+s*L1)*(r3+s*L3)*(D2 pow 2) +
	(r2+s*L2)*(r3+s*L3)*(D3 pow 2)))`,
SFG_TAC
THEN SIMP_TAC [cell3_help1;cell3_help2]
THEN SIMP_TAC [cell3_help3;cell3_help4]
THEN COMPLEX_DIV_TAC);; 



let cell3_help5 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\  
	  ~((r3+s*L3)=Cx(&0)) /\  ~((Cx(&1)+s*R*C)=Cx(&0)) /\
~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r3+s*L3) + D2 pow 2 / 
(r2+s*L2) + D3 pow 2 / (r1+s*L1))) = Cx (&0)) ==> 
((Cx (&1) * Cx (&1) / (r1 + s * L1) * Cx (&1)) *
      (Cx (&1) +
       Cx (&0) +
       Cx (&0) +
       (Cx (&0) +
        Cx (&0) +
        Cx (&0) -
        Cx (&1) / (r3 + s * L3) * D1 * R / (Cx (&1) + s * R * C) * --D1) -
       Cx (&1) / (r2 + s * L2) * D2 * R / (Cx (&1) + s * R * C) * --D2) +
      (Cx (&1) *
       Cx (&1) / (r1 + s * L1) *
       D3 *
       R / (Cx (&1) + s * R * C) *
       --D2 *
       Cx (&1) / (r2 + s * L2) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      (Cx (&1) *
       Cx (&1) / (r1 + s * L1) *
       D3 *
       R / (Cx (&1) + s * R * C) *
       --D1 *
       Cx (&1) / (r3 + s * L3) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r2 + s * L2) * Cx (&1)) *
      (Cx (&1) +
       Cx (&0) +
       Cx (&0) +
       (Cx (&0) +
        Cx (&0) +
        Cx (&0) -
        Cx (&1) / (r3 + s * L3) * D1 * R / (Cx (&1) + s * R * C) * --D1) -
       Cx (&1) / (r1 + s * L1) * D3 * R / (Cx (&1) + s * R * C) * --D3) +
      (Cx (&1) *
       Cx (&1) / (r2 + s * L2) *
       D2 *
       R / (Cx (&1) + s * R * C) *
       --D3 *
       Cx (&1) / (r1 + s * L1) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      (Cx (&1) *
       Cx (&1) / (r2 + s * L2) *
       D2 *
       R / (Cx (&1) + s * R * C) *
       --D1 *
       Cx (&1) / (r3 + s * L3) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      (Cx (&1) *
       Cx (&1) / (r3 + s * L3) *
       D1 *
       R / (Cx (&1) + s * R * C) *
       --D2 *
       Cx (&1) / (r2 + s * L2) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      (Cx (&1) *
       Cx (&1) / (r3 + s * L3) *
       D1 *
       R / (Cx (&1) + s * R * C) *
       --D3 *
       Cx (&1) / (r1 + s * L1) *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r3 + s * L3) * Cx (&1)) *
      (Cx (&1) +
       Cx (&0) +
       Cx (&0) +
       (Cx (&0) +
        Cx (&0) +
        Cx (&0) -
        Cx (&1) / (r2 + s * L2) * D2 * R / (Cx (&1) + s * R * C) * --D2) -
       Cx (&1) / (r1 + s * L1) * D3 * R / (Cx (&1) + s * R * C) * --D3) +
      Cx (&0)) = (((r1+s*L1)*(r2+s*L2)+(r1+s*L1)*(r3+s*L3)+
	  (r2+s*L2)*(r3+s*L3))*(Cx(&1)+s*R*C)+R*((r1+s*L1)*
	  (D1-D2) pow 2 + (r2+s*L2)* (D1-D3) pow 2 + (r3+s*L3)* (D2-D3) pow 2))
/((r1+s*L1)*(r2+s*L2)*(r3+s*L3)*(Cx(&1)+s*R*C))`, CONV_TAC COMPLEX_FIELD);;


let cell3_help6 = 
      prove (`~(r1 + s * L1 = Cx (&0)) /\
 ~(r2 + s * L2 = Cx (&0)) /\
 ~(r3 + s * L3 = Cx (&0)) /\
 ~(Cx (&1) + s * R * C = Cx (&0)) /\
 ~((Cx (&1) + s * R * C) +
   R *
   (D1 pow 2 / (r3 + s * L3) +
    D2 pow 2 / (r2 + s * L2) +
    D3 pow 2 / (r1 + s * L1)) =
   Cx (&0))
 ==> (((r1 + s * L1) * (r2 + s * L2) +
       (r1 + s * L1) * (r3 + s * L3) +
       (r2 + s * L2) * (r3 + s * L3)) *
      (Cx (&1) + s * R * C) +
      R *
      ((r1 + s * L1) * (D1 - D2) pow 2 +
       (r2 + s * L2) * (D1 - D3) pow 2 +
       (r3 + s * L3) * (D2 - D3) pow 2))/
     ((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3) * (Cx (&1) + s * R * C)) /
     (((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1))) /
      (Cx (&1) + s * R * C)) = (((r1 + s * L1) * (r2 + s * L2) +
       (r1 + s * L1) * (r3 + s * L3) +
       (r2 + s * L2) * (r3 + s * L3)) *
      (Cx (&1) + s * R * C) +
      R *
      ((r1 + s * L1) * (D1 - D2) pow 2 +
       (r2 + s * L2) * (D1 - D3) pow 2 +
       (r3 + s * L3) * (D2 - D3) pow 2)) /
     (((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3)) *((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1)))) `, CONV_TAC COMPLEX_FIELD);; 

let cell3_circuit2 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\ 
	  ~((r3+s*L3)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) /\
~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r3+s*L3) + D2 pow 2 / 
(r2+s*L2) + D3 pow 2 /(r1+s*L1))) = Cx (&0)) ==> SFG_Main 
([1,Cx(&1),2;1,Cx(&1),4;1,Cx(&1),7;2,Cx(&1)/(r1+s*L1),3; 
3,Cx(&1),5;3,D3,9;4,Cx(&1)/(r2+s*L2),6;6,Cx(&1),5;6,D2,9;
7,Cx(&1)/(r3+s*L3),8;8,D1,9;8,Cx(&1),5;9,R/(Cx(&1)+s*R*C),10;
10,--D2,4;10,--D3,2;10,--D1,7], 10, 5,1) = (((r1+s*L1) * (r2+s*L2) + 
(r1+s*L1) * (r3+s*L3) + (r2+s*L2) * (r3+s*L3))*(Cx(&1)+s*R*C) + R *
((r1+s*L1) * (D1 - D2) pow 2 + (r2+s*L2) * (D1 - D3) pow 2
 + (r3+s*L3) * (D2 - D3) pow 2)) /
 ((r1+s*L1) * (r2+s*L2) * (r3+s*L3)*(Cx(&1)+s*R*C) + 
 R * ((r1+s*L1) * (r2+s*L2) * D1 pow 2 + 
 (r1+s*L1) * (r3+s*L3) * D2 pow 2 + 
 (r2+s*L2) * (r3+s*L3) * D3 pow 2))`,
SFG_TAC
THEN SIMP_TAC [cell3_help2;cell3_help5]
THEN SIMP_TAC [cell3_help6;cell3_help3]
THEN COMPLEX_DIV_TAC);; 






let cell3_help7 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\  
	  ~((r3+s*L3)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) /\
~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r3+s*L3) + D2 pow 2 
/ (r2+s*L2) + D3 pow 2 / (r1+s*L1))) = Cx (&0)) ==> 
 ((Cx (&1) * Cx (&1) / (r1 + s * L1) * D3) * (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r2 + s * L2) * D2) * (Cx (&1) + Cx (&0)) +
      (Cx (&1) * Cx (&1) / (r3 + s * L3) * D1) * (Cx (&1) + Cx (&0)) +
      Cx (&0)) /
     (((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1))) /
      (Cx (&1) + s * R * C)) = ((r2+s*L2)*(r3+s*L3)*D3  +
  (r1+s*L1)*(r3+s*L3)*D2 +
  (r1+s*L1)*(r2+s*L2)*D1) /((r1+s*L1)*(r2+s*L2)*(r3+s*L3))/
     (((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1))) /
      (Cx (&1) + s * R * C))`, CONV_TAC COMPLEX_FIELD);;


let cell3_help8 = 
      prove (`~(r1 + s * L1 = Cx (&0)) /\
 ~(r2 + s * L2 = Cx (&0)) /\
 ~(r3 + s * L3 = Cx (&0)) /\
 ~(Cx (&1) + s * R * C = Cx (&0)) /\
 ~((Cx (&1) + s * R * C) +
   R *
   (D1 pow 2 / (r3 + s * L3) +
    D2 pow 2 / (r2 + s * L2) +
    D3 pow 2 / (r1 + s * L1)) =
   Cx (&0))
 ==> ((r2 + s * L2) * (r3 + s * L3) * D3 +
      (r1 + s * L1) * (r3 + s * L3) * D2 +
      (r1 + s * L1) * (r2 + s * L2) * D1) /
     ((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3)) /
     (((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1))) /
      (Cx (&1) + s * R * C)) = ((r2 + s * L2) * (r3 + s * L3) * D3 +
      (r1 + s * L1) * (r3 + s * L3) * D2 +
      (r1 + s * L1) * (r2 + s * L2) * D1) * (Cx (&1) + s * R * C) /
 ( ((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3)) * ((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1)))) `, CONV_TAC COMPLEX_FIELD);; 


let cell3_help9 = 
      prove (`~(r1 + s * L1 = Cx (&0)) /\
 ~(r2 + s * L2 = Cx (&0)) /\
 ~(r3 + s * L3 = Cx (&0)) /\
 ~(Cx (&1) + s * R * C = Cx (&0)) /\
 ~((Cx (&1) + s * R * C) +
   R *
   (D1 pow 2 / (r3 + s * L3) +
    D2 pow 2 / (r2 + s * L2) +
    D3 pow 2 / (r1 + s * L1)) =
   Cx (&0))
 ==> 
     (((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3)) *
      ((Cx (&1) + s * R * C) +
       R *
       (D1 pow 2 / (r3 + s * L3) +
        D2 pow 2 / (r2 + s * L2) +
        D3 pow 2 / (r1 + s * L1)))) =    
	((r1 + s * L1) * (r2 + s * L2) * (r3 + s * L3)*(Cx (&1) + s * R * C) +
       R *
       ((r1 + s * L1) * (r2 + s * L2)* D1 pow 2  +
         (r1 + s * L1)  * (r3 + s * L3) * D2 pow 2  +
        (r2 + s * L2) * (r3 + s * L3) * D3 pow 2))  `, CONV_TAC COMPLEX_FIELD);; 




let cell3_circuit3 = 
      prove (`~((r1+s*L1)=Cx(&0)) /\ ~((r2+s*L2)=Cx(&0)) /\  
	  ~((r3+s*L3)=Cx(&0)) /\ ~((Cx(&1)+s*R*C)=Cx(&0)) /\
~(((Cx(&1)+s*R*C) + R * (D1 pow 2 / (r3+s*L3) + D2 pow 2 
/ (r2+s*L2) + D3 pow 2 / (r1+s*L1))) = Cx (&0)) ==> SFG_Main 
([1,Cx(&1),2;1,Cx(&1),4;1,Cx(&1),7;2,Cx(&1)/(r1+s*L1),3;
 3,Cx(&1),5;3,D3,9;4,Cx(&1)/(r2+s*L2),6;6,Cx(&1),5;6,D2,9;
 7,Cx(&1)/(r3+s*L3),8;8,D1,9;8,Cx(&1),5;9,R/(Cx(&1)+s*R*C),10;
 10,--D2,4;10,--D3,2;10,--D1,7], 10, 9,1) = 
((r2+s*L2)*(r3+s*L3)*D3  +
  (r1+s*L1)*(r3+s*L3)*D2 +
  (r1+s*L1)*(r2+s*L2)*D1)*(Cx(&1)+s*R*C)/
     ((r1+s*L1) * (r2+s*L2) * (r3+s*L3)*(Cx(&1)+s*R*C) +
      R * ((r1+s*L1) * (r2+s*L2) * D1 pow 2 + (r1+s*L1) * 
	(r3+s*L3) * D2 pow 2 + (r2+s*L2) * (r3+s*L3) * D3 pow 2))`,
SFG_TAC
THEN SIMP_TAC [cell3_help7;cell3_help2]
THEN SIMP_TAC [cell3_help8;cell3_help9]);; 

(*----------------------------------------------------------------*) 
(*-----------------DC-DC Push-Pull Converter  --------------------*)  
(*----------------------------------------------------------------*) 

(*----------------------------------------------------------------*) 
(*-----------------PWM Circuit Without Controller-----------------*)  
(*----------------------------------------------------------------*) 
let PWM_help1 = 
      prove (`~(C*L=Cx(&0)) /\  ~(s=Cx(&0)) ==> (Cx (&1) +
      Cx (&0) +
      Cx (&0) +
      (Cx (&0) + Cx (&0) + Cx (&0) - Cx (&1) / s * --(r + rc) / L) -
      Cx (&1) / s * Cx (&1) / C * Cx (&1) / s * --Cx (&1) / L)
      = (L*C * s pow 2 + (r+rc)*C*s + Cx (&1))/(L*C* s pow 2)`,
	  CONV_TAC COMPLEX_FIELD);;


let PWM = prove( `~(C*L=Cx(&0)) /\  ~(s=Cx(&0)) ==>  
SFG_Main ([1,Cx(&1)/L,2;2,Cx(&1)/s,3; 3,--(r+rc)/L,2;3,--rc,4;3,Cx(&1)/C,6;
5,rc,4;5,--Cx(&1)/C,6;5,rc/L,2;6,Cx(&1)/s,7;7,--Cx(&1)/L,2;7,Cx(&1),4], 7, 4,1) =
(Cx (&1) - rc* C* s ) / (L*C * s pow 2 + (r+rc)*C*s + Cx (&1))`, SFG_TAC
THEN SIMP_TAC [PWM_help1]
THEN CONV_TAC COMPLEX_FIELD);;


let PWM_SFG = define `PWM_SFG L r rc C s = 
([1,Cx(&1)/L,2;2,Cx(&1)/s,3; 3,--(r+rc)/L,2;3,--rc,4;
3,Cx(&1)/C,6;5,rc,4;5,--Cx(&1)/C,6;5,rc/L,2;6,Cx(&1)/s,7;
7,--Cx(&1)/L,2;7,Cx(&1),4], 7, 4,1)`;;


let NUM_OF_PWM  = prove (` ! L r rc C z. ~(C*L=Cx(&0)) 
/\  ~(z=Cx(&0)) ==> NUMERATOR (PWM_SFG L r rc C z)  =  
	((Cx(&1)-z*C*rc))/(L*C*z pow 2)`,
REWRITE_TAC [PWM_SFG] THEN 
NUMERATOR_TAC
THEN CONV_TAC COMPLEX_FIELD);;




let PWM_resonance_conditions = prove(
` (norm ((Cx (&1))/(C*rc)) <= &1) /\
	(~(C*L = Cx(&0))) /\ (~(rc = Cx(&0))) 
	==> is_resonant (\z. PWM_SFG L r rc C z) z 
	[(Cx(&1))/(C*rc)]`,
	REWRITE_TAC [is_resonant;poles_in_unit_circle;
	are_valid_zeros;in_unit_circle;
	is_valid_zero;set_of_zeros;NUM_OF_PWM] THEN 
	REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC [] 
	THEN REPEAT (POP_ASSUM MP_TAC) 
	THEN REWRITE_TAC[IN_ELIM_THM]
THEN REPEAT (STRIP_TAC)
THEN SUBGOAL_THEN `~(Cx (&1) / (C * rc)= Cx (&0))` ASSUME_TAC
 THEN REPEAT (POP_ASSUM MP_TAC)
THEN TRY(CONV_TAC COMPLEX_FIELD)
THEN REPEAT (POP_ASSUM MP_TAC) 
THEN SIMP_TAC[NUM_OF_PWM]
THEN CONV_TAC COMPLEX_FIELD);;


let PWM_help2 = prove( `  
z = (--((r + rc) * C) + csqrt (((r + rc) * C) pow 2 - 
(Cx(&4)) * (C * L)*(Cx (&1))))/((Cx(&2)) * (C * L)) /\ 
~((C * L) = Cx (&0)) ==> 
(C * L) * z pow 2 + ((r + rc) * C) * z + Cx (&1) = Cx (&0)`,SIMP_TAC[help0] );;

let PWM_help3 = prove( `  
z = (--((r + rc) * C) - csqrt (((r + rc) * C) pow 2 - 
(Cx(&4)) * (C * L) * Cx (&1) ))/((Cx(&2)) * (C * L)) /\
 ~((C * L) = Cx (&0)) ==> 
(C * L) * z pow 2 + ((r + rc) * C) * z + Cx (&1) = Cx (&0)`,SIMP_TAC[help01] );;

let PWM_help4 = prove( `  ~((C * L) = Cx (&0)) ==> 
	(C * L) * ((--((r + rc) * C) + csqrt (((r + rc) * C) pow 2 - 
	(Cx(&4)) * (C * L)*Cx(&1)))/((Cx(&2)) * (C * L) ))  pow 2 + 
	((r + rc) * C) * ((--((r + rc) * C) + csqrt (((r + rc) * C) pow 2 - 
	(Cx(&4)) * (C * L)*Cx(&1)))/((Cx(&2)) * (C * L)))  + Cx(&1) = Cx (&0) /\
(C * L) * ((--((r + rc) * C) - csqrt (((r + rc) * C) pow 2 - 
(Cx(&4)) * (C * L)*Cx(&1)))/((Cx(&2)) * (C * L)))  pow 2 + 
((r + rc) * C) * ((--((r + rc) * C) - csqrt (((r + rc) * C) pow 2 - 
(Cx(&4)) * (C * L)* Cx(&1)))/((Cx(&2)) * (C * L)))  + Cx (&1)  = Cx (&0)`,
SIMP_TAC[PWM_help2;PWM_help3] );;



 let PWM_help5 = prove( `  ~(C *L = Cx (&0)) ==> 
	((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      (--(r + rc) * C +
       csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
      (Cx (&2) * C * L) *
      (r + rc) *
      C +
      Cx (&1)  = (C * L) *
         ((--((r + rc) * C) +
           csqrt
           (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
          (Cx (&2) * C * L)) pow
         2 +
         ((r + rc) * C) *
         (--((r + rc) * C) +
          csqrt
          (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
         (Cx (&2) * C * L) +
         Cx (&1)  /\
     (((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1) )) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      (--(r + rc) * C -
       csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1) )) /
      (Cx (&2) * C * L) *
      (r + rc) *
      C +
      Cx (&1)) = (C * L) *
         ((--((r + rc) * C) -
           csqrt
           (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
          (Cx (&2) * C * L)) pow
         2 +
         ((r + rc) * C) *
         (--((r + rc) * C) -
          csqrt
          (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1) )) /
         (Cx (&2) * C * L) +
     Cx (&1) `,SIMP_TAC [COMPLEX_MUL_LNEG] THEN CONV_TAC COMPLEX_FIELD );;

let PWM_help6  = prove( `~(C*L=Cx(&0)) ==> ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C +  Cx (&1)  =
    Cx (&0) /\
    ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C + Cx (&1) =
    Cx (&0)`, SIMP_TAC [PWM_help5 ;PWM_help4 ] THEN CONV_TAC COMPLEX_FIELD) ;;


let PWM_help7  = prove( `~(C*L=Cx(&0)) ==>  ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C +  Cx (&1) = ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C +  Cx (&1) /\ ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C + Cx (&1) = ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L)*Cx (&1))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C + Cx (&1)`, SIMP_TAC [COMPLEX_MUL_RID] ) ;;


let PWM_help8  = prove( `~(C*L=Cx(&0)) ==> ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C +
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C +  Cx (&1)  =
    Cx (&0) /\
    ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      ((--(r + rc) * C -
        csqrt (((r + rc) * C) pow 2 - Cx (&4) * (C * L))) /
       (Cx (&2) * C * L))  *
      (r + rc) *
      C + Cx (&1) =
    Cx (&0)`,  SIMP_TAC [PWM_help6;PWM_help7] ) ;;

let DENOM_OF_PWM = 
      prove (` !L r rc C z. ~(C*L=Cx(&0)) /\  ~(z=Cx(&0)) ==> 
	  DENOMINATOR (PWM_SFG L r rc C z) = 
	  (z pow 2 *C *L + z *(r+rc)*C + Cx(&1))/(L*C*z pow 2)`,
REWRITE_TAC [PWM_SFG] THEN 
DENOMINATOR_TAC
THEN CONV_TAC COMPLEX_FIELD);;
	
let PWM_stability_conditions = prove( `
(norm ((--(r+rc)*C + csqrt (((r+rc)* C) pow 2 - 
(Cx(&4))*(C*L)))/((Cx(&2)) * C*L)) <= &1) 
/\ (norm ((--(r+rc)*C - csqrt (((r+rc)* C) pow 2 - 
(Cx(&4))*(C*L)))/((Cx(&2)) * C*L)) <= &1) 
 /\ ~(C*L=Cx(&0)) /\  ~((--(r+rc)*C + csqrt (((r+rc)* C) pow 2 - 
 (Cx(&4))*(C*L)))/((Cx(&2)) * C*L)=Cx(&0)) 
/\  ~((--(r+rc)*C - csqrt (((r+rc)* C) pow 2 -
 (Cx(&4))*(C*L)))/((Cx(&2)) * C*L)=Cx(&0)) 
	==> is_stable (\z. PWM_SFG L r rc C z) z 
	[(--(r+rc)*C + csqrt (((r+rc)* C) pow 2 - 
	(Cx(&4))*(C*L)))/((Cx(&2)) * C*L);
	(--(r+rc)*C - csqrt (((r+rc)* C) pow 2 - 
	(Cx(&4))*(C*L)))/((Cx(&2)) * C*L)]`,
REWRITE_TAC [is_stable;poles_in_unit_circle;
	are_valid_poles;in_unit_circle;
	is_valid_pole;set_of_poles;DENOM_OF_PWM] THEN 
	REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC) THEN
REWRITE_TAC[IN_ELIM_THM] 
THEN SIMP_TAC[DENOM_OF_PWM]
THEN SIMP_TAC[PWM_help8]
THEN TRY(CONV_TAC COMPLEX_FIELD)
THEN REWRITE_TAC[IN_ELIM_THM]
THEN SIMP_TAC[DENOM_OF_PWM]
THEN SIMP_TAC[PWM_help8]
THEN CONV_TAC COMPLEX_FIELD);;


(*----------------------------------------------------------------*) 
(*-----------------PWM Circuit With Controller--------------------*)  
(*----------------------------------------------------------------*) 



let PWM2_help1 = 
      prove (`~(C=Cx(&0)) /\ ~(L=Cx(&0)) /\  ~(s=Cx(&0)) ==> 
(Cx (&1) +
  Cx (&0) +
  Cx (&0) +
  (Cx (&0) +
   Cx (&0) +
   (Cx (&0) +
    Cx (&0) +
    (Cx (&0) + Cx (&0) + Cx (&0) - Cx (&1) / s * --(r + rc) / L) -
    Cx (&1) / s * Cx (&1) / C * Cx (&1) / s * --Cx (&1) / L) -
   Cx (&1) / L * Cx (&1) / s * --ki) -
  Cx (&1) / L * Cx (&1) / s * Cx (&1) / C * Cx (&1) / s * --kv)
= (s pow 3 *C *L + s pow 2 *(r+rc+ki)*C + 
(Cx(&1)+kv)*s)/(L*C* s pow 3)`, CONV_TAC COMPLEX_FIELD);;

let PWM2_help2 = 
      prove (`~(C=Cx(&0)) /\ ~(L=Cx(&0)) /\  ~(s=Cx(&0)) /\
 ~(s pow 3 * C * L + s pow 2 * (r + rc + ki) * C + (Cx (&1) + kv) * s =
   Cx (&0)) ==> 
((ky / s * Cx (&1) / L * Cx (&1) / s * --rc) * (Cx (&1) + Cx (&0)) +
      (ky / s *
       Cx (&1) / L *
       Cx (&1) / s *
       Cx (&1) / C *
       Cx (&1) / s *
       Cx (&1)) *
      (Cx (&1) + Cx (&0)) +
      Cx (&0)) /
     ((s * s * s * C * L + s * s * (r + rc + ki) * C + s * (Cx (&1) + kv)) /
      (s * s * s * L * C)) = ky*(Cx (&1)-rc*C*s)/(s * s * s * C * L + 
	s * s * (r + rc + ki) * C + s * (Cx (&1) + kv))`, CONV_TAC COMPLEX_FIELD);;


let PWM2 = prove( `~(C*L=Cx(&0)) /\  ~(s=Cx(&0)) /\ 
~((s pow 3 *C *L + s pow 2 *(r+rc+ki)*C + (Cx(&1)+kv)*s) = Cx(&0)) 
==>  SFG_Main ([1,ky/s,2;2,Cx(&1)/L,3;3,Cx(&1)/s,4; 4,--(r+rc)/L,3;
4,--ki,2;4,--rc,5;4,Cx(&1)/C,6;6,Cx(&1)/s,8;7,--Cx(&1)/C,6;7,rc/L,3;
7,rc,4;8,--kv,2;8,--Cx(&1)/L,3;8,Cx(&1),5], 8, 5,1) = 
(ky*(Cx(&1)-s*C*rc))/(s pow 3 *C *L + s pow 2 *(r+rc+ki)*C + 
(Cx(&1)+kv)*s)`, SFG_TAC
THEN SIMP_TAC [PWM2_help1;PWM2_help2]
THEN CONV_TAC COMPLEX_FIELD);;


let PWM2_SFG = define `PWM2_SFG ky L r rc C kv ki s = 
([1,ky/s,2;2,Cx(&1)/L,3;3,Cx(&1)/s,4; 4,--(r+rc)/L,3;
4,--ki,2;4,--rc,5;4,Cx(&1)/C,6;6,Cx(&1)/s,8;7,--Cx(&1)/C,6;
7,rc/L,3;7,rc,4;8,--kv,2;8,--Cx(&1)/L,3;8,Cx(&1),5], 8, 5,1)`;;


let NUM_OF_PWM2 =  prove (` !ky L r rc C kv ki z.  
~(rc=Cx(&0)) /\ ~(C*L=Cx(&0)) /\  ~(z=Cx(&0)) ==> 
NUMERATOR (PWM2_SFG ky L r rc C kv ki z) =  (ky*(Cx(&1)-z*C*rc))/(L*C*z pow 3)`,
REWRITE_TAC [PWM2_SFG] THEN 
NUMERATOR_TAC
THEN CONV_TAC COMPLEX_FIELD);;

let PWM2_resonance_condition = 
prove( ` (norm (Cx(&1)/(C*rc)) <= &1) 
/\ ~(C*L=Cx(&0)) /\ ~(rc=Cx(&0))
	==> is_resonant (\z. PWM2_SFG ky L r rc C kv ki z) z  
	[Cx(&1)/(C*rc)]`,
REWRITE_TAC [is_resonant;poles_in_unit_circle;
	are_valid_zeros;in_unit_circle;
	is_valid_zero;set_of_zeros;NUM_OF_PWM2]
THEN REWRITE_TAC[IN_ELIM_THM] THEN STRIP_TAC
THEN MP_TAC (COMPLEX_FIELD `~(C * L = Cx (&0)) /\
 ~(rc = Cx (&0)) ==> ~ ((Cx (&1) / (C * rc)) = (Cx(&0))) `)
THEN REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC[NUM_OF_PWM2]
THEN CONV_TAC COMPLEX_FIELD);;



let DENOM_OF_PWM2 = 
      prove (` !ky L r rc C kv ki z. ~(C*L=Cx(&0)) /\  ~(z=Cx(&0))
	  ==> DENOMINATOR (PWM2_SFG ky L r rc C kv ki z) =  
	  (z pow 3 *C *L + z pow 2 *(r+rc+ki)*C + (Cx(&1)+kv)*z)/(L*C*z pow 3)`,
REWRITE_TAC [PWM2_SFG] THEN 
DENOMINATOR_TAC
THEN CONV_TAC COMPLEX_FIELD);;
	


let PWM2_help3 = prove( `  z = (--((r + rc + ki) * C) - csqrt (((r + rc + ki) * C) pow 2 - 
(Cx(&4)) * (C * L) * (Cx(&1)+kv) ))/((Cx(&2)) * (C * L)) /\ ~((C * L) = Cx (&0)) ==> 
	(C * L) * z pow 2 + ((r + rc + ki) * C) * z + (Cx(&1)+kv) = Cx (&0)`,SIMP_TAC[help01] );;

let PWM2_help4 = prove( `  z = (--((r + rc + ki) * C) + csqrt (((r + rc + ki) * C) pow 2 - 
(Cx(&4)) * (C * L) * (Cx(&1)+kv) ))/((Cx(&2)) * (C * L)) /\ ~((C * L) = Cx (&0)) ==> 
	(C * L) * z pow 2 + ((r + rc + ki) * C) * z + (Cx(&1)+kv) = Cx (&0)`,SIMP_TAC[help0] );;





let PWM2_help5  = prove( `  ~((C * L) = Cx (&0)) ==> 
	(C * L) * ((--((r + rc + ki) * C) + csqrt (((r + rc + ki) * C) pow 2 - 
(Cx(&4)) * (C * L) * (Cx (&1) + kv)))/((Cx(&2)) * (C * L)))  pow 2 + 
((r + rc + ki) * C) * ((--((r + rc + ki) * C) + csqrt (((r + rc + ki) * C) pow 2 - 
(Cx(&4)) * (C * L) * (Cx (&1) + kv)))/((Cx(&2)) * (C * L)))  + (Cx (&1) + kv) = Cx (&0) /\
(C * L) * ((--((r + rc + ki) * C) - csqrt (((r + rc + ki) * C) pow 2 - 
(Cx(&4)) * (C * L) * (Cx (&1) + kv)))/((Cx(&2)) * (C * L)))  pow 2 + 
((r + rc + ki) * C) * ((--((r + rc + ki) * C) - csqrt (((r + rc + ki) * C) pow 2 - 
(Cx(&4)) * (C * L) * (Cx (&1) + kv)))/((Cx(&2)) * (C * L)))  + (Cx (&1) + kv) = Cx (&0)
`,SIMP_TAC[PWM2_help4;PWM2_help3] );;


 let PWM2_help6 = prove( `  ~(C *L = Cx (&0)) ==> 
	((--(r + rc + ki) * C +
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) pow
     3 *
     C *
     L +
     ((--(r + rc + ki) * C +
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) pow
     2 *
     (r + rc + ki) *
     C +
     (Cx (&1) + kv) *
     (--(r + rc + ki) * C +
      csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
     (Cx (&2) * C * L)  = (((--(r + rc + ki) * C +
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) pow
     2 *
     C *
     L +
     ((--(r + rc + ki) * C +
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L))  *
     (r + rc + ki) *
     C +
     (Cx (&1) + kv)) *
     (--(r + rc + ki) * C +
      csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
     (Cx (&2) * C * L) /\
((--(r + rc + ki) * C -
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) pow
     3 *
     C *
     L +
     ((--(r + rc + ki) * C -
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) pow
     2 *
     (r + rc + ki) *
     C +
     (Cx (&1) + kv) *
     (--(r + rc + ki) * C -
      csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
     (Cx (&2) * C * L)  = (((--(r + rc + ki) * C -
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) pow
     2 *
     C *
     L +
     ((--(r + rc + ki) * C -
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L))  *
     (r + rc + ki) *
     C +
     (Cx (&1) + kv)) *
     (--(r + rc + ki) * C -
      csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
     (Cx (&2) * C * L) `,CONV_TAC COMPLEX_FIELD );;


let PWM2_help7 = prove( `  ~(C *L = Cx (&0)) ==> 
	(((--(r + rc + ki) * C +
        csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      (--(r + rc + ki) * C +
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L) *
      (r + rc + ki) *
      C +
      Cx (&1) +
      kv) = (C * L) *
         ((--((r + rc + ki) * C) +
           csqrt
           (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
          (Cx (&2) * C * L)) pow
         2 +
         ((r + rc + ki) * C) *
         (--((r + rc + ki) * C) +
          csqrt
          (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
         (Cx (&2) * C * L) +
         Cx (&1) +
         kv /\
(((--(r + rc + ki) * C -
        csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
       (Cx (&2) * C * L)) pow
      2 *
      C *
      L +
      (--(r + rc + ki) * C -
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L) *
      (r + rc + ki) *
      C +
      Cx (&1) +
      kv) = (C * L) *
         ((--((r + rc + ki) * C) -
           csqrt
           (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
          (Cx (&2) * C * L)) pow
         2 +
         ((r + rc + ki) * C) *
         (--((r + rc + ki) * C) -
          csqrt
          (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
         (Cx (&2) * C * L) +
         Cx (&1) +
         kv `,SIMP_TAC [COMPLEX_MUL_LNEG] THEN CONV_TAC COMPLEX_FIELD );;

let PWM2_help8 = prove( `~(C*L=Cx(&0)) ==> (((--(r + rc + ki) * C +
        csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
       (Cx (&2) * C * L)) pow
      3 *
      C *
      L +
      ((--(r + rc + ki) * C +
        csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
       (Cx (&2) * C * L)) pow
      2 *
      (r + rc + ki) *
      C +
      (Cx (&1) + kv) *
      (--(r + rc + ki) * C +
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) =
    Cx (&0) /\
(((--(r + rc + ki) * C -
        csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
       (Cx (&2) * C * L)) pow
      3 *
      C *
      L +
      ((--(r + rc + ki) * C -
        csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
       (Cx (&2) * C * L)) pow
      2 *
      (r + rc + ki) *
      C +
      (Cx (&1) + kv) *
      (--(r + rc + ki) * C -
       csqrt (((r + rc + ki) * C) pow 2 - Cx (&4) * (C * L) * (Cx (&1) + kv))) /
      (Cx (&2) * C * L)) =
    Cx (&0)`, SIMP_TAC [PWM2_help7;PWM2_help6;PWM2_help5] THEN CONV_TAC COMPLEX_FIELD) ;;



let PWM2_stability_condition = 
prove( `(norm ((--(r+rc+ki)*C + csqrt (((r+rc+ki)* C) pow 2 - 
(Cx(&4))*(C*L)*(Cx(&1) + kv)))/((Cx(&2)) * C*L)) <= &1) 
/\ (norm ((--(r+rc+ki)*C - csqrt (((r+rc+ki)* C) pow 2 -
 (Cx(&4))*(C*L)*(Cx(&1) + kv)))/((Cx(&2)) * C*L)) <= &1) 
 /\ ~(C*L=Cx(&0)) /\  ~((--(r+rc+ki)*C + 
csqrt (((r+rc+ki)* C) pow 2 - (Cx(&4))*(C*L)*(Cx(&1) + kv)))/
((Cx(&2)) * C*L)=Cx(&0)) 
/\  ~((--(r+rc+ki)*C - csqrt (((r+rc+ki)* C) pow 2 - 
(Cx(&4))*(C*L)*(Cx(&1) + kv)))/((Cx(&2)) * C*L)=Cx(&0)) 
	==> is_stable (\z. PWM2_SFG ky L r rc C kv ki z) z 
	[(--(r+rc+ki)*C + csqrt (((r+rc+ki)* C) pow 2 - 
(Cx(&4))*(C*L)*(Cx(&1) + kv)))/((Cx(&2)) * C*L);
	(--(r+rc+ki)*C - csqrt (((r+rc+ki)* C) pow 2 - 
(Cx(&4))*(C*L)*(Cx(&1) + kv)))/((Cx(&2)) * C*L)]`,
REWRITE_TAC [is_stable;poles_in_unit_circle;
	are_valid_poles;in_unit_circle;
	is_valid_pole;set_of_poles;DENOM_OF_PWM2] THEN 
	REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN 
	REPEAT (POP_ASSUM MP_TAC)  THEN
REWRITE_TAC[IN_ELIM_THM] THEN 
SIMP_TAC[DENOM_OF_PWM2] THEN
SIMP_TAC[PWM2_help8] THEN
TRY (CONV_TAC COMPLEX_FIELD) THEN
REWRITE_TAC[IN_ELIM_THM] THEN
SIMP_TAC[DENOM_OF_PWM2] THEN
SIMP_TAC[PWM2_help8] THEN CONV_TAC COMPLEX_FIELD );;



(*----------------------------------------------------------------*) 
(*--------------------------Mulit legged Robot--------------------*)  
(*----------------------------------------------------------------*) 
 
let Legged_Robot = 
	prove (`! (G1:complex) G2 G3 G4 G5 G6 G7 G8 H2 H3  H6 H7.  
   SFG_Main ([1,G1,2;1,G5,5;2,G2,3;3,H2,2;3,G3,4;4,H3,3;4,G4,8;5,G6,6; 
   6,H6,5;6,G7,7;7,H7,6;7,G8,8], 8, 8,1) =  
(G1*G2*G3*G4*(Cx(&1) - G6*H6 - G7*H7) +  
G5*G6*G7*G8*(Cx(&1) - G2*H2 - G3*H3)) / 
(Cx(&1) - G2*H2 - G3*H3 - G6*H6 - G7*H7 + G2*H2*G6*H6 + G2*H2*G7*H7 +  
 G3*H3*G6*H6 + G3*H3*G7*H7)`,
SFG_TAC
THEN COMPLEX_DIV_TAC);; 
