(* ========================================================================= *)
(*                                                                           *)
(*                        Mason_Formula.ml                                   *)
(*                                                                           *)
(*        (c) Copyright, Sidi Mohamed BEILLAHI, Muhammad Umair Siddique      *)
(*                       2015                                                *)
(*                       Hardware Verification Group,                        *)
(*                       Concordia University                                *)
(*                                                                           *)
(*     Contact: <beillahi@encs.concordia.ca>,                                *)
(*       	<muh_sidd@ece.concordia.ca>                                 *)
(*                                                                           *)
(* Last update: March 02, 2015                                               *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)


needs "SFG.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)
(*--------------------**********************------------------*)
let GAIN = define 
	`!(t:path).
		    (GAIN [] = Cx(&0)) /\ 
                    (GAIN (CONS h t) =
                     if (t = []) then snd_of_trpl h else
                    (snd_of_trpl h) * (GAIN t))`;;
		
let LEFT = define 
	`!(t:path) (a:branch).
    	  (LEFT a [] = F ) /\
    	  (LEFT a (CONS b t) = 
     	  if ((lst_of_trpl a = fst_of_trpl b) \/
              (fst_of_trpl a = lst_of_trpl b) \/
              (lst_of_trpl a = lst_of_trpl b) \/
              (fst_of_trpl a = fst_of_trpl b)) then T
          else LEFT a t)`;;
		  
let RIGHT = define 
	`!(t:path) (b:branch).
    	  (RIGHT [] b = F ) /\
    	  (RIGHT (CONS a t) b = 
     	  if ((lst_of_trpl b = fst_of_trpl a) \/
              (fst_of_trpl b = lst_of_trpl a) \/
              (lst_of_trpl b = lst_of_trpl a) \/
              (fst_of_trpl b = fst_of_trpl a)) then T
          else RIGHT t b)`;;

let IS_TOUCHING = define 
		`!(t1: path) (t2: path).
            (IS_TOUCHING [] [] = F) /\ 
     		(IS_TOUCHING [] t1 = F) /\ 
			(IS_TOUCHING t2 [] = F) /\ 
			(IS_TOUCHING (CONS a t1) (CONS b t2) = 
			if ((RIGHT (CONS a t1) b) \/ (LEFT a (CONS b t2))) 
			then T else IS_TOUCHING t1 t2)`;;

			
			
			
			
let TOUCHING_LOOP = define 
		`!(t1:path) (t:(path)list).
               (TOUCHING_LOOP t1 [] =  []) /\
			   (TOUCHING_LOOP  t1 (CONS t2 t) =
                        if (IS_TOUCHING t1 t2) then (TOUCHING_LOOP t1 t)
                        else CONS t2 (TOUCHING_LOOP t1 t))`;;

let TOUCHING_LOOP_MAIN = define 
			`!(t:(path)list).
			(TOUCHING_LOOP_MAIN [] = []) /\
                    (TOUCHING_LOOP_MAIN (CONS h t) = 
                   (CONS (TOUCHING_LOOP h t) (TOUCHING_LOOP_MAIN t)))`;;

								 	
let GAIN_PATH_LISTPATHS  = define 
			`!(t1: path) (t:(path)list). 
    			(GAIN_PATH_LISTPATHS t1 [] = Cx(&0))
    			/\ (GAIN_PATH_LISTPATHS t1 (CONS t2 t) =
    		(((GAIN t1) * (GAIN t2)) + (GAIN_PATH_LISTPATHS t1 t)))`;;
	 
							 
let ARE_SAME_PATH_def = define 
			`!(t1: path) (t2: path). 
    		      (ARE_SAME_PATH [] [] = T ) /\ 
		      (ARE_SAME_PATH (CONS a t1) [] = F ) /\ 
		      (ARE_SAME_PATH [] (CONS b t2) = F ) /\
		      (ARE_SAME_PATH (CONS a t1) (CONS b t2) = 
		       if((fst_of_trpl a = fst_of_trpl b) /\ 
		       (lst_of_trpl a= lst_of_trpl b)) then 
		       (T /\ (ARE_SAME_PATH t1 t2)) 
		        else F)`;;
							
let LEFT_OF_LIST_def = define 
			`!(t1: path) (t: (path) list).
     		  (LEFT_OF_LIST t1 [] = F ) /\ 
                  (LEFT_OF_LIST t1 (CONS t2 t) 
		 = if (ARE_SAME_PATH t1 t2) then T 
                   else (LEFT_OF_LIST t1 t))`;;
 

let COMMON_ELES_LISTS = define 
			`!(t1: (path) list) (t2: (path) list).
     		    (COMMON_ELES_LISTS [] [] = []) /\ 
                (COMMON_ELES_LISTS [] t1 = [])
                /\ (COMMON_ELES_LISTS (CONS a t1) t2 =	
		   if (LEFT_OF_LIST a t2) then (CONS a 
		   (COMMON_ELES_LISTS t1 t2))
                   else (COMMON_ELES_LISTS t1 t2))`;;


									  
									  
let EXTRACT_ELE_PLACE = define 
		`!(n:num) (t1: path) (t: (path) list).
    		(EXTRACT_ELE_PLACE n t1 (CONS t2 t) = 
		 if (ARE_SAME_PATH t1 t2) then n 
                else EXTRACT_ELE_PLACE (n+1) t1 t) `;;		
				
let LISTS_ARE_EMPTY = define 
		`!(t: ((path) list) list). 
    		(LISTS_ARE_EMPTY [] = T) /\ 
		    (LISTS_ARE_EMPTY (CONS t1 t) =			    
		    if (t1=[]) then (LISTS_ARE_EMPTY t) else F )` ;;

					   
					   
let PRODUCT_THREE_MORE = define  
	`!(t1: path) (t3: (path) list) (t: ((path) list) list) (l: (path) list) 
			 (tl: ((path) list) list). 
     	(PRODUCT_THREE_MORE t1 (CONS t2 t3) [] l tl = Cx(&0)) /\ 
        (PRODUCT_THREE_MORE t1 [] t l tl = Cx(&0)) /\ 
        (PRODUCT_THREE_MORE t1 (CONS t2 t3) (CONS t4 t) l tl = 
               		 if(LISTS_ARE_EMPTY (CONS t4 t)) then Cx(&0) 
               		 else ((PRODUCT_THREE_MORE t1 t3 t l tl) - 
			 ((GAIN t1) * ((GAIN_PATH_LISTPATHS t2 
			 (COMMON_ELES_LISTS t3 (EL 
			 (EXTRACT_ELE_PLACE 0 t2 l) tl))) +  
			 (PRODUCT_THREE_MORE t2 
			 (COMMON_ELES_LISTS t4 t3) t l tl)))))` ;;



let DELTA_MINUS = define 
	`!(t2: (path) list) (t: ((path) list) list) (l: (path) list) 
			  (tl: ((path) list) list).
		(DELTA_MINUS [] [] l tl = Cx(&0)) /\ 
        (DELTA_MINUS (CONS t1 t2) [] l tl = Cx(&0)) /\  
		(DELTA_MINUS [] (CONS t3 t) l tl = Cx(&0)) /\ 
		(DELTA_MINUS (CONS t1 t2) (CONS t3 t) l tl  = 
               ((GAIN_PATH_LISTPATHS t1 t3) + 
			   (PRODUCT_THREE_MORE t1 t3 t l tl) + 
			   (DELTA_MINUS t2 t l tl) - (GAIN t1))) ` ;;

let DETERMINANT = define 
	`!(t: (path) list). 
    	DETERMINANT t = 
		(Cx(&1) + (DELTA_MINUS t (TOUCHING_LOOP_MAIN t) t 
		(TOUCHING_LOOP_MAIN t))) `;;

		
		
		
		
let TOUCHING_LOOPS_FORWARD = define 
	`!(f: path) (t: (path) list).
    	(TOUCHING_LOOPS_FORWARD f t = 
		(TOUCHING_LOOP_MAIN (TOUCHING_LOOP f t))) ` ;;

let FORWARD_DELTA_MINUS = define 
		`!(f: path) (t: (path) list). 
    		FORWARD_DELTA_MINUS f t = 
    		    (DELTA_MINUS (TOUCHING_LOOP f t) 
				(TOUCHING_LOOPS_FORWARD f t) 
				(TOUCHING_LOOP f t) 
				(TOUCHING_LOOPS_FORWARD f t)) `;;

let PRODUCT_FORWARD_DELTA = define 
		`!(l: (path) list) (t: (path) list).
    	(PRODUCT_FORWARD_DELTA [] t = Cx(&0)) /\ 
    	(PRODUCT_FORWARD_DELTA (CONS f l) t =
				 (((GAIN f)*(Cx(&1) + 
				 (FORWARD_DELTA_MINUS f t))) + 
				 (PRODUCT_FORWARD_DELTA l t))) ` ;;



let MASON_GAIN = define 
		`!(l: (path) list) (t: (path) list).
    	       (MASON_GAIN l t = 
    	       ((PRODUCT_FORWARD_DELTA l t) / (DETERMINANT t)))` ;;

(*--------------------------------------------------------------------*)
let SFG_Main_AID = define `!(sfg: SFG) (P: (num) list) (H: ((num) list) list) 
	(G: ((num) list) list).
	SFG_Main_AID sfg P H G =  MASON_GAIN (FC_REWRITING 
	(FORWARD_CIRCUIT 0 (EC_replace_vector_head P (lst_of_four sfg)) 
	  H G (thd_of_four sfg) (snd_of_four sfg)) (fst_of_four sfg))
	  (EC_REWRITING (EC_ALL_POINT 0 (EC_replace_vector_head P 1) 
	   H G (snd_of_four sfg) (snd_of_four sfg) P) (fst_of_four sfg))`;;

let SFG_Main = define `!(sfg: SFG).
	SFG_Main sfg
	= if((fst_of_four sfg)=[]) then Cx(&0) else  
	(SFG_Main_AID sfg (EC_create_vector (snd_of_four sfg)) 
	(EC_create_matrix (snd_of_four sfg)) 
	(EC_G_creation 0 (EC_start_nodes (fst_of_four sfg)) 
	(EC_end_nodes (fst_of_four sfg)) (EC_create_vector (snd_of_four sfg)))) ` ;;




(*---------------------------------------------------------------------*)
(*----------------------MASON_GAIN Simplication Tactic ----------------*)
(*---------------------------------------------------------------------*)
(*----- Its takes a list of theorems "xs" and simplifies the goal -----*)
(*---------------------------------------------------------------------*)
(*---------------------------------------------------------------------*)


let MASON_SIMP_TAC =

REWRITE_TAC [MASON_GAIN] THEN
REWRITE_TAC [PRODUCT_FORWARD_DELTA] THEN
REWRITE_TAC [FORWARD_DELTA_MINUS; GAIN ] THEN
REWRITE_TAC [RIGHT; LEFT;IS_TOUCHING; TOUCHING_LOOPS_FORWARD] THEN
REWRITE_TAC [DETERMINANT; DELTA_MINUS; 
             PRODUCT_THREE_MORE] THEN 
REWRITE_TAC [LISTS_ARE_EMPTY; COMMON_ELES_LISTS] THEN
REWRITE_TAC [snd_of_trpl ; fst_of_trpl;lst_of_trpl ] THEN
REWRITE_TAC [TOUCHING_LOOP_MAIN; TOUCHING_LOOP; GAIN_PATH_LISTPATHS] THEN
REWRITE_TAC [ARE_SAME_PATH_def; LEFT_OF_LIST_def;EL;
             EXTRACT_ELE_PLACE] THEN 
REWRITE_TAC [NOT_CONS_NIL] THEN 
CONV_TAC NUM_REDUCE_CONV THEN NUMBER_SFG_TAC1[HD;TL;ADD;ADD_CLAUSES;EL] THEN
NUMBER_SFG_TAC2[HD;TL;EL] THEN CONV_TAC NUM_REDUCE_CONV;;


let COMPLEX_DIV_TAC = 
    REPEAT STRIP_TAC THEN REWRITE_TAC[complex_div] THEN TRY (BINOP_TAC) THEN 
    TRY (CONV_TAC COMPLEX_FIELD) THEN 
    TRY (AP_TERM_TAC) THEN 
    TRY (CONV_TAC COMPLEX_FIELD);;
	
	
let ECFEC_SIMP_TAC1 = 
 REWRITE_TAC [SFG_Main;component_of_four;NOT_CONS_NIL]
THEN NUMBER_SFG_TAC1 [minus; EC_H_MODIFIE1; 
LENGTH;EC_H_MODIFIE2;EC_H_MODIFIE; EC_H_MODIFIE0; 
EC_remove_node;EC_add_node;  
EC_TEST; IN_LIST; EECTEST;   
EC_add_one_to_list_head; EC_create_matrix; EC_replace_vector_head; 
EC_create_vector; EC_start_nodes; EC_end_nodes;HD; 
TL;component_of_four;EL;APPEND]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [minus;HD;TL;component_of_four;EL;APPEND;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [minus; EC_H_MODIFIE1;LENGTH; 
EC_H_MODIFIE2;EC_H_MODIFIE; EC_H_MODIFIE0; EC_remove_node; 
EC_add_node;EC_TEST; IN_LIST; EECTEST;   
EC_add_one_to_list_head; EC_create_matrix; EC_replace_vector_head; 
EC_create_vector; EC_start_nodes; EC_end_nodes;HD; 
TL;component_of_trpl;EL;APPEND;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN 
(REPEAT (CHANGED_TAC TAC_EC_G_creation)) THEN
NUMBER_SFG_TAC1 [SFG_Main_AID;
EC_add_one_to_list_head;HD;component_of_four] THEN  
NUMBER_SFG_TAC2 [SFG_Main_AID;HD;component_of_four]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [minus] THEN  
NUMBER_SFG_TAC2 [minus_one_aid]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [SFG_Main_AID;EL;minus;HD;TL;EECTEST;FST; 
SND;EC_replace_vector_head;APPEND;component_of_four] THEN  
NUMBER_SFG_TAC2 [SFG_Main_AID;EL;minus;HD;TL;EECTEST;FST; 
SND;EC_replace_vector_head;APPEND;component_of_four] THEN  
CONV_TAC NUM_REDUCE_CONV 
THEN ONCE_REWRITE_TAC [ EC_ALL_POINT]  
THEN REWRITE_TAC [EC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL] THEN  
CONV_TAC NUM_REDUCE_CONV THEN 
NUMBER_SFG_TAC1 [EL;minus;HD;TL] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;ECSKIP] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL; 
EC_next_node;IN_LIST] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL; 
EC_next_node;IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;;


let ECFEC_SIMP_TAC2 = 
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
CONV_TAC NUM_REDUCE_CONV 
THEN NUMBER_SFG_TAC1 [ EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND]   
THEN CONV_TAC NUM_REDUCE_CONV THEN   
REWRITE_TAC [NOT_CONS_NIL] THEN 
NUMBER_SFG_TAC1 [EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND; 
REPLICATE] THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [ EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND; 
REPLICATE] THEN REWRITE_TAC [NOT_CONS_NIL] THEN 
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EC_REWRITING; EC_REWRITING0; 
EC_REWRITING1; EC_REWRITING2;HD] THEN  
NUMBER_SFG_TAC2 [HD]  THEN REWRITE_TAC [NOT_CONS_NIL] THEN  
CONV_TAC NUM_REDUCE_CONV 
THEN ONCE_REWRITE_TAC [FORWARD_CIRCUIT] 
THEN REWRITE_TAC [FC_next_node_main] THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;REPLICATE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;FC_next_node; 
REPLICATE] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;FC_next_node; 
REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
REWRITE_TAC [NOT_IN_LIST] THEN  
CONV_TAC NUM_REDUCE_CONV ;;


let ECFEC_SIMP_TAC3 = 
ONCE_REWRITE_TAC [FC_FORWARD] THEN  
NUMBER_SFG_TAC1 [minus; FCTEST; FCTEST1; HD;TL;component_of_trpl;EL;APPEND]   
THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [ minus;HD;TL;component_of_trpl;EL;APPEND;REPLICATE]  
THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [FCTEST; FCTEST1; minus; HD;TL;component_of_trpl; 
EL;APPEND;REPLICATE] THEN CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
NUMBER_SFG_TAC2 [EL;minus;HD;TL;EECTEST;component_of_trpl
;APPEND;EC_add_node] THEN  
CONV_TAC NUM_REDUCE_CONV 
THEN NUMBER_SFG_TAC1 [ FC_REWRITING; FC_REWRITING1; 
FC_REWRITING0; EC_REWRITING2;HD;TL;component_of_trpl;
EL;APPEND]  THEN  
CONV_TAC NUM_REDUCE_CONV THEN  REWRITE_TAC [NOT_CONS_NIL] THEN 
NUMBER_SFG_TAC1 [FC_REWRITING; FC_REWRITING1; 
FC_REWRITING0; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND;REPLICATE] THEN  
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [ FC_REWRITING; FC_REWRITING0; 
FC_REWRITING1; EC_REWRITING2;HD;TL;component_of_trpl
;EL;APPEND;REPLICATE] THEN  
REWRITE_TAC [NOT_CONS_NIL] THEN 
CONV_TAC NUM_REDUCE_CONV THEN  
NUMBER_SFG_TAC1 [FC_REWRITING; FC_REWRITING0; 
FC_REWRITING1; EC_REWRITING2;HD] THEN  
NUMBER_SFG_TAC2 [HD]  THEN  
REWRITE_TAC [NOT_CONS_NIL] THEN CONV_TAC NUM_REDUCE_CONV     
THEN (REPEAT (CHANGED_TAC MASON_SIMP_TAC))  ;;

(*-----------------------------------------------------------------*)
let SFG_TAC = ECFEC_SIMP_TAC1
THEN (REPEAT (CHANGED_TAC TAC_EC))
THEN ECFEC_SIMP_TAC2
THEN (REPEAT (CHANGED_TAC TAC_FC))
THEN ECFEC_SIMP_TAC3 ;; 
